@@ -1,31 +1,27 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from collections import defaultdict
-from itertools import tee, zip_longest
+from decimal import Decimal
 
 try:
     import pygal
 except ImportError:
     pygal = None
 from dateutil.relativedelta import relativedelta
-from sql import Null, Literal, Column, With
-from sql.aggregate import Sum, Max, Min, Count
+from sql import Column, Literal, Null, Union, With
+from sql.aggregate import Count, Max, Min, Sum
 from sql.conditionals import Coalesce
-from sql.functions import CurrentTimestamp, DateTrunc, Power, Ceil, Log
+from sql.functions import Ceil, CurrentTimestamp, DateTrunc, Log, Power
+from sql.operators import Concat
 
-from trytond.pool import Pool
-from trytond.model import ModelSQL, ModelView, UnionMixin, fields
-from trytond.tools import grouped_slice, reduce_ids
-from trytond.transaction import Transaction
-from trytond.pyson import Eval, If
-from trytond.wizard import Wizard, StateTransition, StateAction
 from trytond.i18n import lazy_gettext
-
-
-def pairwise(iterable):
-    a, b = tee(iterable)
-    next(b, None)
-    return zip_longest(a, b)
+from trytond.model import ModelSQL, ModelView, UnionMixin, fields, sum_tree
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
+from trytond.pyson import Eval, If
+from trytond.tools import grouped_slice, pairwise_longest, reduce_ids
+from trytond.transaction import Transaction
+from trytond.wizard import StateAction, StateTransition, Wizard
 
 
 class Abstract(ModelSQL):
@@ -34,10 +30,9 @@ class Abstract(ModelSQL):
         'company.company', lazy_gettext("sale.msg_sale_reporting_company"))
     number = fields.Integer(lazy_gettext("sale.msg_sale_reporting_number"),
         help=lazy_gettext("sale.msg_sale_reporting_number_help"))
-    revenue = fields.Numeric(
+    revenue = Monetary(
         lazy_gettext("sale.msg_sale_reporting_revenue"),
-        digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
+        digits='currency', currency='currency')
     revenue_trend = fields.Function(
         fields.Char(lazy_gettext("sale.msg_sale_reporting_revenue_trend")),
         'get_trend')
@@ -47,10 +42,6 @@ class Abstract(ModelSQL):
             'currency.currency',
             lazy_gettext("sale.msg_sale_reporting_currency")),
         'get_currency')
-    currency_digits = fields.Function(
-        fields.Integer(
-            lazy_gettext("sale.msg_sale_reporting_currency_digits")),
-        'get_currency_digits')
 
     @classmethod
     def table_query(cls):
@@ -60,18 +51,56 @@ class Abstract(ModelSQL):
             group_by=cls._group_by(tables, withs),
             with_=withs.values())
 
+    @classmethod
+    def _sale_line(cls, length, index, company_id=None):
+        pool = Pool()
+        Line = pool.get('sale.line')
+        Sale = pool.get('sale.sale')
+
+        line = Line.__table__()
+        sale = Sale.__table__()
+
+        return (line
+            .join(sale, condition=line.sale == sale.id)
+            .select(
+                (line.id * length + index).as_('id'),
+                *cls._sale_line_columns(line, sale),
+                where=sale.state.in_(cls._sale_states())
+                & (sale.company == company_id),
+                ))
+
+    @classmethod
+    def _sale_line_columns(cls, line, sale):
+        return [
+            line.product.as_('product'),
+            Coalesce(line.actual_quantity, line.quantity).as_('quantity'),
+            line.unit_price.as_('unit_price'),
+            Concat('sale.sale,', line.sale).as_('order'),
+            sale.sale_date.as_('date'),
+            sale.company.as_('company'),
+            sale.currency.as_('currency'),
+            sale.party.as_('customer'),
+            sale.warehouse.as_('location'),
+            sale.shipment_address.as_('shipment_address'),
+            ]
+
+    @classmethod
+    def _lines(cls):
+        return [cls._sale_line]
+
     @classmethod
     def _joins(cls):
         pool = Pool()
         Company = pool.get('company.company')
         Currency = pool.get('currency.currency')
-        Line = pool.get('sale.line')
-        Sale = pool.get('sale.sale')
+        context = Transaction().context
 
         tables = {}
-        tables['line'] = line = Line.__table__()
-        tables['line.sale'] = sale = Sale.__table__()
-        tables['line.sale.company'] = company = Company.__table__()
+        company = context.get('company')
+        lines = cls._lines()
+        tables['line'] = line = Union(*(
+                l(len(lines), i, company) for i, l in enumerate(lines)))
+        tables['line.company'] = company = Company.__table__()
         withs = {}
         currency_sale = With(query=Currency.currency_rate_sql())
         withs['currency_sale'] = currency_sale
@@ -79,32 +108,29 @@ class Abstract(ModelSQL):
         withs['currency_company'] = currency_company
 
         from_item = (line
-            .join(sale, condition=line.sale == sale.id)
             .join(currency_sale,
-                condition=(sale.currency == currency_sale.currency)
-                & (currency_sale.start_date <= sale.sale_date)
+                condition=(line.currency == currency_sale.currency)
+                & (currency_sale.start_date <= line.date)
                 & ((currency_sale.end_date == Null)
-                    | (currency_sale.end_date > sale.sale_date))
+                    | (currency_sale.end_date > line.date))
                 )
-            .join(company, condition=sale.company == company.id)
+            .join(company, condition=line.company == company.id)
             .join(currency_company,
                 condition=(company.currency == currency_company.currency)
-                & (currency_company.start_date <= sale.sale_date)
+                & (currency_company.start_date <= line.date)
                 & ((currency_company.end_date == Null)
-                    | (currency_company.end_date > sale.sale_date))
+                    | (currency_company.end_date > line.date))
                 ))
         return from_item, tables, withs
 
     @classmethod
     def _columns(cls, tables, withs):
         line = tables['line']
-        sale = tables['line.sale']
         currency_company = withs['currency_company']
         currency_sale = withs['currency_sale']
 
-        quantity = Coalesce(line.actual_quantity, line.quantity)
         revenue = cls.revenue.sql_cast(
-            Sum(quantity * line.unit_price
+            Sum(line.quantity * line.unit_price
                 * currency_company.rate / currency_sale.rate))
         return [
             cls._column_id(tables, withs).as_('id'),
@@ -112,9 +138,9 @@ class Abstract(ModelSQL):
             CurrentTimestamp().as_('create_date'),
             cls.write_uid.sql_cast(Literal(Null)).as_('write_uid'),
             cls.write_date.sql_cast(Literal(Null)).as_('write_date'),
-            sale.company.as_('company'),
+            line.company.as_('company'),
             revenue.as_('revenue'),
-            Count(sale.id, distinct=True).as_('number'),
+            Count(line.order, distinct=True).as_('number'),
             ]
 
     @classmethod
@@ -124,25 +150,30 @@ class Abstract(ModelSQL):
 
     @classmethod
     def _group_by(cls, tables, withs):
-        sale = tables['line.sale']
-        return [sale.company]
+        line = tables['line']
+        return [line.company]
 
     @classmethod
     def _where(cls, tables, withs):
+        pool = Pool()
+        Location = pool.get('stock.location')
         context = Transaction().context
-        sale = tables['line.sale']
+        line = tables['line']
 
-        where = sale.company == context.get('company')
-        where &= sale.state.in_(cls._sale_states())
+        where = Literal(True)
         from_date = context.get('from_date')
         if from_date:
-            where &= sale.sale_date >= from_date
+            where &= line.date >= from_date
         to_date = context.get('to_date')
         if to_date:
-            where &= sale.sale_date <= to_date
+            where &= line.date <= to_date
         warehouse = context.get('warehouse')
         if warehouse:
-            where &= sale.warehouse == warehouse
+            locations = Location.search([
+                    ('parent', 'child_of', warehouse),
+                    ],
+                query=True)
+            where &= line.location.in_(locations)
         return where
 
     @classmethod
@@ -152,7 +183,7 @@ class Abstract(ModelSQL):
     @property
     def time_series_all(self):
         delta = self._period_delta()
-        for ts, next_ts in pairwise(self.time_series or []):
+        for ts, next_ts in pairwise_longest(self.time_series or []):
             yield ts
             if delta and next_ts:
                 date = ts.date + delta
@@ -180,13 +211,10 @@ class Abstract(ModelSQL):
     def get_currency(self, name):
         return self.company.currency.id
 
-    def get_currency_digits(self, name):
-        return self.company.currency.digits
-
 
 class AbstractTimeseries(Abstract):
 
-    date = fields.Date("Date")
+    date = fields.Date(lazy_gettext('sale.msg_sale_reporting_date'))
 
     @classmethod
     def __setup__(cls):
@@ -201,8 +229,8 @@ class AbstractTimeseries(Abstract):
     @classmethod
     def _column_date(cls, tables, withs):
         context = Transaction().context
-        sale = tables['line.sale']
-        date = DateTrunc(context.get('period'), sale.sale_date)
+        line = tables['line']
+        date = DateTrunc(context.get('period'), line.date)
         date = cls.date.sql_cast(date)
         return date
 
@@ -222,15 +250,13 @@ class Context(ModelView):
             If(Eval('to_date') & Eval('from_date'),
                 ('from_date', '<=', Eval('to_date')),
                 ()),
-            ],
-        depends=['to_date'])
+            ])
     to_date = fields.Date("To Date",
         domain=[
             If(Eval('from_date') & Eval('to_date'),
                 ('to_date', '>=', Eval('from_date')),
                 ()),
-            ],
-        depends=['from_date'])
+            ])
     period = fields.Selection([
             ('year', "Year"),
             ('month', "Month"),
@@ -273,6 +299,29 @@ class Context(ModelView):
         return Transaction().context.get('warehouse')
 
 
+class Main(Abstract, ModelView):
+    "Sale Reporting"
+    __name__ = 'sale.reporting.main'
+
+    time_series = fields.Function(fields.One2Many(
+            'sale.reporting.main.time_series', None,
+            lazy_gettext('sale.msg_sale_reporting_time_series')),
+        'get_time_series')
+
+    def get_rec_name(self, name):
+        return ''
+
+    def get_time_series(self, name):
+        pool = Pool()
+        Timeseries = pool.get('sale.reporting.main.time_series')
+        return [t.id for t in Timeseries.search([])]
+
+
+class MainTimeseries(AbstractTimeseries, ModelView):
+    "Sale Reporting"
+    __name__ = 'sale.reporting.main.time_series'
+
+
 class CustomerMixin(object):
     __slots__ = ()
     customer = fields.Many2One(
@@ -280,30 +329,35 @@ class CustomerMixin(object):
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
 
     @classmethod
     def _columns(cls, tables, withs):
-        sale = tables['line.sale']
+        line = tables['line']
         return super(CustomerMixin, cls)._columns(tables, withs) + [
-            sale.party.as_('customer')]
+            line.customer.as_('customer')]
 
     @classmethod
     def _group_by(cls, tables, withs):
-        sale = tables['line.sale']
+        line = tables['line']
         return super(CustomerMixin, cls)._group_by(tables, withs) + [
-            sale.party]
+            line.customer]
 
     def get_rec_name(self, name):
         return self.customer.rec_name
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('customer.rec_name', *clause[1:])]
+
 
 class Customer(CustomerMixin, Abstract, ModelView):
     "Sale Reporting per Customer"
     __name__ = 'sale.reporting.customer'
 
     time_series = fields.One2Many(
-        'sale.reporting.customer.time_series', 'customer', "Time Series")
+        'sale.reporting.customer.time_series', 'customer',
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
 
     @classmethod
     def __setup__(cls):
@@ -312,8 +366,15 @@ class Customer(CustomerMixin, Abstract, ModelView):
 
     @classmethod
     def _column_id(cls, tables, withs):
-        sale = tables['line.sale']
-        return sale.party
+        line = tables['line']
+        return line.customer
+
+    @classmethod
+    def _where(cls, tables, withs):
+        line = tables['line']
+        where = super()._where(tables, withs)
+        where &= line.customer != Null
+        return where
 
 
 class CustomerTimeseries(CustomerMixin, AbstractTimeseries, ModelView):
@@ -321,6 +382,181 @@ class CustomerTimeseries(CustomerMixin, AbstractTimeseries, ModelView):
     __name__ = 'sale.reporting.customer.time_series'
 
 
+class CustomerCategoryMixin:
+    __slots__ = ()
+    category = fields.Many2One('party.category', "Category")
+
+    @classmethod
+    def _joins(cls):
+        pool = Pool()
+        PartyCategory = pool.get('party.party-party.category')
+        from_item, tables, withs = super()._joins()
+        if 'line.customer.party_category' not in tables:
+            party_category = PartyCategory.__table__()
+            tables['line.customer.party_category'] = party_category
+            line = tables['line']
+            from_item = (from_item
+                .join(party_category,
+                    condition=line.customer == party_category.party))
+        return from_item, tables, withs
+
+    @classmethod
+    def _columns(cls, tables, withs):
+        party_category = tables['line.customer.party_category']
+        return super()._columns(tables, withs) + [
+            party_category.category.as_('category')]
+
+    @classmethod
+    def _column_id(cls, tables, withs):
+        pool = Pool()
+        Category = pool.get('party.category')
+        category = Category.__table__()
+        line = tables['line']
+        party_category = tables['line.customer.party_category']
+        # Get a stable number of categories over time
+        # by using a number one order bigger.
+        nb_category = category.select(
+            Power(10, (Ceil(Log(Max(category.id))) + Literal(1))))
+        return Min(line.id * nb_category + party_category.id)
+
+    @classmethod
+    def _group_by(cls, tables, withs):
+        party_category = tables['line.customer.party_category']
+        return super()._group_by(tables, withs) + [party_category.category]
+
+    @classmethod
+    def _where(cls, tables, withs):
+        party_category = tables['line.customer.party_category']
+        where = super()._where(tables, withs)
+        where &= party_category.category != Null
+        return where
+
+    def get_rec_name(self, name):
+        return self.category.rec_name if self.category else None
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('category.rec_name', *clause[1:])]
+
+
+class CustomerCategory(CustomerCategoryMixin, Abstract, ModelView):
+    "Sale Reporting per Customer Category"
+    __name__ = 'sale.reporting.customer.category'
+
+    time_series = fields.One2Many(
+        'sale.reporting.customer.category.time_series', 'category',
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('category', 'ASC'))
+
+    @classmethod
+    def _column_id(cls, tables, withs):
+        party_category = tables['line.customer.party_category']
+        return party_category.category
+
+
+class CustomerCategoryTimeseries(
+        CustomerCategoryMixin, AbstractTimeseries, ModelView):
+    "Sale Reporting per Customer Category"
+    __name__ = 'sale.reporting.customer.category.time_series'
+
+
+class CustomerCategoryTree(ModelSQL, ModelView):
+    "Sale Reporting per Customer Category"
+    __name__ = 'sale.reporting.customer.category.tree'
+
+    name = fields.Function(
+        fields.Char("Name"), 'get_name', searcher='search_name')
+    parent = fields.Many2One('sale.reporting.customer.category.tree', "Parent")
+    children = fields.One2Many(
+        'sale.reporting.customer.category.tree', 'parent', "Children")
+    revenue = fields.Function(Monetary(
+            "Revenue", digits='currency', currency='currency'), 'get_total')
+
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"), 'get_currency')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('name', 'ASC'))
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Category = pool.get('party.category')
+        return Category.__table__()
+
+    @classmethod
+    def get_name(cls, categories, name):
+        pool = Pool()
+        Category = pool.get('party.category')
+        categories = Category.browse(categories)
+        return {c.id: c.name for c in categories}
+
+    @classmethod
+    def search_name(cls, name, clause):
+        pool = Pool()
+        Category = pool.get('party.category')
+        return [('id', 'in', Category.search([clause], query=True))]
+
+    @classmethod
+    def order_name(cls, tables):
+        pool = Pool()
+        Category = pool.get('party.category')
+        table, _ = tables[None]
+        if 'category' not in tables:
+            category = Category.__table__()
+            tables['category'] = {
+                None: (category, table.id == category.id),
+                }
+        return Category.name.convert_order(
+                'name', tables['category'], Category)
+
+    def time_series_all(self):
+        return []
+
+    @classmethod
+    def get_total(cls, categories, names):
+        pool = Pool()
+        ReportingCustomerCategory = pool.get(
+            'sale.reporting.customer.category')
+        reporting_product_category = ReportingCustomerCategory.__table__()
+        cursor = Transaction().connection.cursor()
+
+        categories = cls.search([
+                ('parent', 'child_of', [c.id for c in categories]),
+                ])
+        ids = [c.id for c in categories]
+        reporting_customer_categories = []
+        for sub_ids in grouped_slice(ids):
+            sub_ids = list(sub_ids)
+            where = reduce_ids(reporting_product_category.id, sub_ids)
+            cursor.execute(
+                *reporting_product_category.select(
+                    reporting_product_category.id, where=where))
+            reporting_customer_categories.extend(r for r, in cursor)
+
+        result = {}
+        reporting_product_categories = ReportingCustomerCategory.browse(
+            reporting_customer_categories)
+        for name in names:
+            values = defaultdict(Decimal,
+                {c.id: getattr(c, name) for c in reporting_product_categories})
+            result[name] = sum_tree(categories, values)
+        return result
+
+    def get_currency(self, name):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company = Transaction().context.get('company')
+        if company:
+            return Company(company).currency.id
+
+
 class ProductMixin(object):
     __slots__ = ()
     product = fields.Many2One(
@@ -328,7 +564,7 @@ class ProductMixin(object):
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
 
     @classmethod
     def _columns(cls, tables, withs):
@@ -352,13 +588,18 @@ class ProductMixin(object):
     def get_rec_name(self, name):
         return self.product.rec_name if self.product else None
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('product.rec_name', *clause[1:])]
+
 
 class Product(ProductMixin, Abstract, ModelView):
     "Sale Reporting per Product"
     __name__ = 'sale.reporting.product'
 
     time_series = fields.One2Many(
-        'sale.reporting.product.time_series', 'product', "Time Series")
+        'sale.reporting.product.time_series', 'product',
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
 
     @classmethod
     def __setup__(cls):
@@ -376,21 +617,21 @@ class ProductTimeseries(ProductMixin, AbstractTimeseries, ModelView):
     __name__ = 'sale.reporting.product.time_series'
 
 
-class CategoryMixin(object):
+class ProductCategoryMixin(object):
     __slots__ = ()
     category = fields.Many2One(
         'product.category', "Category",
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
 
     @classmethod
     def _joins(cls):
         pool = Pool()
         Product = pool.get('product.product')
         TemplateCategory = pool.get('product.template-product.category.all')
-        from_item, tables, withs = super(CategoryMixin, cls)._joins()
+        from_item, tables, withs = super()._joins()
         if 'line.product' not in tables:
             product = Product.__table__()
             tables['line.product'] = product
@@ -409,7 +650,7 @@ class CategoryMixin(object):
     @classmethod
     def _columns(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        return super(CategoryMixin, cls)._columns(tables, withs) + [
+        return super()._columns(tables, withs) + [
             template_category.category.as_('category')]
 
     @classmethod
@@ -428,26 +669,30 @@ class CategoryMixin(object):
     @classmethod
     def _group_by(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        return super(CategoryMixin, cls)._group_by(tables, withs) + [
-            template_category.category]
+        return super()._group_by(tables, withs) + [template_category.category]
 
     @classmethod
     def _where(cls, tables, withs):
         template_category = tables['line.product.template_category']
-        where = super(CategoryMixin, cls)._where(tables, withs)
+        where = super()._where(tables, withs)
         where &= template_category.category != Null
         return where
 
     def get_rec_name(self, name):
         return self.category.rec_name if self.category else None
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('category.rec_name', *clause[1:])]
 
-class Category(CategoryMixin, Abstract, ModelView):
-    "Sale Reporting per Category"
-    __name__ = 'sale.reporting.category'
+
+class ProductCategory(ProductCategoryMixin, Abstract, ModelView):
+    "Sale Reporting per Product Category"
+    __name__ = 'sale.reporting.product.category'
 
     time_series = fields.One2Many(
-        'sale.reporting.category.time_series', 'category', "Time Series")
+        'sale.reporting.product.category.time_series', 'category',
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
 
     @classmethod
     def __setup__(cls):
@@ -460,25 +705,26 @@ class Category(CategoryMixin, Abstract, ModelView):
         return template_category.category
 
 
-class CategoryTimeseries(CategoryMixin, AbstractTimeseries, ModelView):
-    "Sale Reporting per Category"
-    __name__ = 'sale.reporting.category.time_series'
+class ProductCategoryTimeseries(
+        ProductCategoryMixin, AbstractTimeseries, ModelView):
+    "Sale Reporting per Product Category"
+    __name__ = 'sale.reporting.product.category.time_series'
 
 
-class CategoryTree(ModelSQL, ModelView):
-    "Sale Reporting per Category"
-    __name__ = 'sale.reporting.category.tree'
+class ProductCategoryTree(ModelSQL, ModelView):
+    "Sale Reporting per Product Category"
+    __name__ = 'sale.reporting.product.category.tree'
 
-    name = fields.Function(fields.Char("Name"), 'get_name')
-    parent = fields.Many2One('sale.reporting.category.tree', "Parent")
+    name = fields.Function(
+        fields.Char("Name"), 'get_name', searcher='search_name')
+    parent = fields.Many2One('sale.reporting.product.category.tree', "Parent")
     children = fields.One2Many(
-        'sale.reporting.category.tree', 'parent', "Children")
-    revenue = fields.Function(
-        fields.Numeric("Revenue", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total')
+        'sale.reporting.product.category.tree', 'parent', "Children")
+    revenue = fields.Function(Monetary(
+            "Revenue", digits='currency'), 'get_total')
 
-    currency_digits = fields.Function(
-        fields.Integer("Currency Digits"), 'get_currency_digits')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"), 'get_currency')
 
     @classmethod
     def __setup__(cls):
@@ -498,6 +744,12 @@ class CategoryTree(ModelSQL, ModelView):
         categories = Category.browse(categories)
         return {c.id: c.name for c in categories}
 
+    @classmethod
+    def search_name(cls, name, clause):
+        pool = Pool()
+        Category = pool.get('product.category')
+        return [('id', 'in', Category.search([clause], query=True))]
+
     @classmethod
     def order_name(cls, tables):
         pool = Pool()
@@ -517,64 +769,39 @@ class CategoryTree(ModelSQL, ModelView):
     @classmethod
     def get_total(cls, categories, names):
         pool = Pool()
-        ReportingCategory = pool.get('sale.reporting.category')
-        table = cls.__table__()
-        reporting_category = ReportingCategory.__table__()
+        ReportingProductCategory = pool.get('sale.reporting.product.category')
+        reporting_product_category = ReportingProductCategory.__table__()
         cursor = Transaction().connection.cursor()
 
         categories = cls.search([
                 ('parent', 'child_of', [c.id for c in categories]),
                 ])
         ids = [c.id for c in categories]
-        parents = {}
-        reporting_categories = []
+        reporting_product_categories = []
         for sub_ids in grouped_slice(ids):
             sub_ids = list(sub_ids)
-            where = reduce_ids(table.id, sub_ids)
-            cursor.execute(*table.select(table.id, table.parent, where=where))
-            parents.update(cursor)
-
-            where = reduce_ids(reporting_category.id, sub_ids)
+            where = reduce_ids(reporting_product_category.id, sub_ids)
             cursor.execute(
-                *reporting_category.select(reporting_category.id, where=where))
-            reporting_categories.extend(r for r, in cursor)
+                *reporting_product_category.select(
+                    reporting_product_category.id, where=where))
+            reporting_product_categories.extend(r for r, in cursor)
 
         result = {}
-        reporting_categories = ReportingCategory.browse(reporting_categories)
+        reporting_product_categories = ReportingProductCategory.browse(
+            reporting_product_categories)
         for name in names:
-            values = dict.fromkeys(ids, 0)
-            values.update(
-                (c.id, getattr(c, name)) for c in reporting_categories)
-            result[name] = cls._sum_tree(categories, values, parents)
+            values = defaultdict(
+                Decimal,
+                {c.id: getattr(c, name) for c in reporting_product_categories})
+            result[name] = sum_tree(categories, values)
         return result
 
-    @classmethod
-    def _sum_tree(cls, categories, values, parents):
-        result = values.copy()
-        categories = set((c.id for c in categories))
-        leafs = categories - set(parents.values())
-        while leafs:
-            for category in leafs:
-                categories.remove(category)
-                parent = parents.get(category)
-                if parent in result:
-                    result[parent] += result[category]
-            next_leafs = set(categories)
-            for category in categories:
-                parent = parents.get(category)
-                if not parent:
-                    continue
-                if parent in next_leafs and parent in categories:
-                    next_leafs.remove(parent)
-            leafs = next_leafs
-        return result
-
-    def get_currency_digits(self, name):
+    def get_currency(self, name):
         pool = Pool()
         Company = pool.get('company.company')
         company = Transaction().context.get('company')
         if company:
-            return Company(company).currency.digits
+            return Company(company).currency.id
 
 
 class CountryMixin(object):
@@ -586,40 +813,150 @@ class CountryMixin(object):
         pool = Pool()
         Address = pool.get('party.address')
         from_item, tables, withs = super(CountryMixin, cls)._joins()
-        if 'line.sale.shipment_address' not in tables:
+        if 'line.shipment_address' not in tables:
             address = Address.__table__()
-            tables['line.sale.shipment_address'] = address
-            sale = tables['line.sale']
+            tables['line.shipment_address'] = address
+            line = tables['line']
             from_item = (from_item
-                .join(address, condition=sale.shipment_address == address.id))
+                .join(address, condition=line.shipment_address == address.id))
         return from_item, tables, withs
 
     @classmethod
     def _columns(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return super(CountryMixin, cls)._columns(tables, withs) + [
             address.country.as_('country')]
 
     @classmethod
     def _group_by(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return super(CountryMixin, cls)._group_by(tables, withs) + [
             address.country]
 
     @classmethod
     def _where(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         where = super(CountryMixin, cls)._where(tables, withs)
         where &= address.country != Null
         return where
 
+    def get_rec_name(self, name):
+        return self.country.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('country.rec_name', *clause[1:])]
+
+
+class RegionTree(ModelSQL, ModelView):
+    "Sale Reporting per Region"
+    __name__ = 'sale.reporting.region.tree'
+
+    name = fields.Function(fields.Char("Name"), 'get_name')
+    parent = fields.Many2One('sale.reporting.region.tree', "Parent")
+    subregions = fields.One2Many(
+        'sale.reporting.region.tree', 'parent', "Subregions")
+
+    revenue = fields.Function(Monetary(
+            "Revenue", digits='currency'), 'get_total')
+
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"), 'get_currency')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('name', 'ASC'))
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Region = pool.get('country.region')
+        return Region.__table__()
+
+    @classmethod
+    def get_name(cls, regions, name):
+        pool = Pool()
+        Region = pool.get('country.region')
+        regions = Region.browse(regions)
+        return {r.id: r.name for r in regions}
+
+    @classmethod
+    def order_name(cls, tables):
+        pool = Pool()
+        Region = pool.get('country.region')
+        table, _ = tables[None]
+        if 'region' not in tables:
+            region = Region.__table__()
+            tables['region'] = {
+                None: (region, table.id == region.id),
+                }
+        return Region.name.convert_order(
+                'name', tables['region'], Region)
+
+    @classmethod
+    def get_total(cls, regions, names):
+        pool = Pool()
+        ReportingCountry = pool.get('sale.reporting.country')
+
+        regions = cls.search([
+                ('parent', 'child_of', [r.id for r in regions]),
+                ])
+        reporting_countries = []
+
+        result = {}
+        reporting_countries = ReportingCountry.search([
+                ('country.region', 'in', [r.id for r in regions]),
+                ])
+        for name in names:
+            values = defaultdict(Decimal)
+            for reporting_country in reporting_countries:
+                values[reporting_country.country.region.id] += (
+                    getattr(reporting_country, name))
+            result[name] = sum_tree(regions, values)
+        return result
+
+    def get_currency(self, name):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company = Transaction().context.get('company')
+        if company:
+            return Company(company).currency.id
+
+
+class OpenRegionTree(Wizard):
+    "Open Region"
+    __name__ = 'sale.reporting.region.tree.open'
+
+    start = StateAction('sale.act_reporting_country_tree')
+
+    def do_start(self, action):
+        pool = Pool()
+        Country = pool.get('country.country')
+        CountryTree = pool.get('sale.reporting.country.tree')
+        countries = Country.search([
+                ('region', 'child_of', [r.id for r in self.records], 'parent'),
+                ])
+        ids = [
+            CountryTree.union_shard(c.id, 'sale.reporting.country')
+            for c in countries]
+        data = {
+            'ids': ids,
+            }
+        name_suffix = ', '.join(r.rec_name for r in self.records[:5])
+        if len(self.records) > 5:
+            name_suffix += ',...'
+        action['name'] += ' (%s)' % name_suffix
+        return action, data
+
 
 class Country(CountryMixin, Abstract):
     "Sale Reporting per Country"
     __name__ = 'sale.reporting.country'
 
     time_series = fields.One2Many(
-        'sale.reporting.country.time_series', 'country', "Time Series")
+        'sale.reporting.country.time_series', 'country',
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
 
     @classmethod
     def __setup__(cls):
@@ -628,12 +965,9 @@ class Country(CountryMixin, Abstract):
 
     @classmethod
     def _column_id(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return address.country
 
-    def get_rec_name(self, name):
-        return self.country.rec_name
-
 
 class CountryTimeseries(CountryMixin, AbstractTimeseries, ModelView):
     "Sale Reporting per Country"
@@ -646,23 +980,30 @@ class SubdivisionMixin(CountryMixin):
 
     @classmethod
     def _columns(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return super(SubdivisionMixin, cls)._columns(tables, withs) + [
             address.subdivision.as_('subdivision')]
 
     @classmethod
     def _group_by(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return super(SubdivisionMixin, cls)._group_by(tables, withs) + [
             address.subdivision]
 
     @classmethod
     def _where(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         where = super(SubdivisionMixin, cls)._where(tables, withs)
         where &= address.subdivision != Null
         return where
 
+    def get_rec_name(self, name):
+        return self.subdivision.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('subdivision.rec_name', *clause[1:])]
+
 
 class Subdivision(SubdivisionMixin, Abstract):
     "Sale Reporting per Subdivision"
@@ -670,7 +1011,7 @@ class Subdivision(SubdivisionMixin, Abstract):
 
     time_series = fields.One2Many(
         'sale.reporting.country.subdivision.time_series', 'subdivision',
-        "Time Series")
+        lazy_gettext('sale.msg_sale_reporting_time_series'))
 
     @classmethod
     def __setup__(cls):
@@ -679,25 +1020,23 @@ class Subdivision(SubdivisionMixin, Abstract):
 
     @classmethod
     def _column_id(cls, tables, withs):
-        address = tables['line.sale.shipment_address']
+        address = tables['line.shipment_address']
         return address.subdivision
 
-    def get_rec_name(self, name):
-        return self.subdivision.rec_name
-
 
 class SubdivisionTimeseries(SubdivisionMixin, AbstractTimeseries, ModelView):
     "Sale Reporting per Subdivision"
     __name__ = 'sale.reporting.country.subdivision.time_series'
 
 
-class Region(UnionMixin, Abstract, ModelView):
-    "Sale Reporting per Region"
-    __name__ = 'sale.reporting.region'
+class CountryTree(UnionMixin, Abstract, ModelView):
+    "Sale Reporting per Country"
+    __name__ = 'sale.reporting.country.tree'
 
     region = fields.Function(fields.Char("Region"), 'get_rec_name')
-    parent = fields.Many2One('sale.reporting.region', "Parent")
-    children = fields.One2Many('sale.reporting.region', 'parent', "Children")
+    parent = fields.Many2One('sale.reporting.country.tree', "Parent")
+    children = fields.One2Many(
+        'sale.reporting.country.tree', 'parent', "Children")
 
     @classmethod
     def union_models(cls):
@@ -705,7 +1044,7 @@ class Region(UnionMixin, Abstract, ModelView):
 
     @classmethod
     def union_column(cls, name, field, table, Model):
-        column = super(Region, cls).union_column(
+        column = super().union_column(
             name, field, table, Model)
         if (name == 'parent'
                 and Model.__name__ == 'sale.reporting.country.subdivision'):
@@ -732,9 +1071,9 @@ class Region(UnionMixin, Abstract, ModelView):
         return record.time_series
 
 
-class OpenRegion(Wizard):
-    "Open Region"
-    __name__ = 'sale.reporting.region.open'
+class OpenCountryTree(Wizard):
+    "Open Country"
+    __name__ = 'sale.reporting.country.tree.open'
 
     start = StateTransition()
     country = StateAction('sale.act_reporting_country_time_series')
